; dircheck/asm - kjw/bqsd - 08/82
;
	TITLE	'<Dos+ II - Directory Checker - A.00>'
;
	ORG	2200H		;high overlay area
;
EVAL	EQU	83		;evaluate input line
VIDKEY	EQU	12		;video/keyboard
OPEN	EQU	40		;open file
VDCHAR	EQU	8		;character to video
KBCHAR	EQU	4		;character from keyboard
ERROR	EQU	39		;display error
HLDKEY	EQU	29		;hold processor
BINDEC	EQU	21		;binary to decimal
SETBRK	EQU	3		;setup break processor
;
LOCDCT	EQU	032DH		;locate drive DCT
UCASE	EQU	031EH
REGSAV	EQU	0303H
TOPMEM	EQU	0173H		;top memory
LOWMEM	EQU	0171H		;bottom memory
BUFFER	EQU	1A00H		;user buffer
STRING	EQU	BUFFER		;use for string
;
DCHECK	EQU	0		;check drive for ready
DREAD	EQU	3		;read sector
SREAD	EQU	6		;read system sector
;
SVC	EQU	8		;execute SVC
DISKIO	EQU	10H		;direct disk I/O
;
ERR8	EQU	8		;drive not ready
ERR19	EQU	19		;invalid filespec
ERR48	EQU	48		;invalid parameter
ERR82	EQU	82		;insufficient memory
ERR86	EQU	86		;terminated
ERR88	EQU	88		;not available
ERR92	EQU	92		;no extent terminator
ERR93	EQU	93		;link does not link back
ERR94	EQU	94		;track beyond disk
ERR95	EQU	95		;bad forward link
ERR96	EQU	96		;bad forward link
ERR99	EQU	99		;EOF corrupt
ERR100	EQU	100		;extension too early
ERR101	EQU	101		;extension not assigned
ERR102	EQU	102		;multiple files > gran
ERR103	EQU	103		;invalid hit
ERR104	EQU	104		;zero hit
ERR105	EQU	105		;extended invalid hit
ERR106	EQU	106		;extended zero hit
;
	PAGE
;
;	program entry point
;
ENTRY	LD	(STACK),SP	;save stack pointer
;
	PUSH	HL		;save command pointer
	LD	HL,HELLO	;sign on message
	CALL	DISPLAY		;display message
;
	LD	HL,0		;turn off break processor
	LD	A,SETBRK	;SVC #
	RST	SVC
	LD	(OLDBRK),HL	;save old vector
;
	LD	A,HLDKEY	;turn on hold processor
	LD	B,0		;turn it off
	RST	SVC
;
	POP	HL		;restore pointer
;
;	evaluate input line
;
SCAN	LD	IX,CHKBLK	;eval block
	LD	A,EVAL		;SVC #
	RST	SVC		;fetch params
	JR	NZ,EXIT		;error exit
;
	LD	A,(IX)		;fetch user params
	AND	6		;any input?
	JR	Z,APARMS	;ask for parameters
;
	LD	DE,DCB		;point to DCB
	LD	A,(DE)		;get type byte
	BIT	6,A		;file here?
	JP	NZ,CHKFILE	;check out the file
	BIT	7,A		;device here?
	LD	B,ERR48		;invalid command
	JR	Z,EXITB		;no drive #
;
	AND	0FH		;get device number
	SUB	8		;must be 8-15
	LD	B,ERR88		;function not available
	JR	C,EXITB		;can't check low devices
;
	LD	(DRIVE),A	;save drive number
	JR	CHKDRIV		;check the drive
;
;	error vector
;
TERMIN	LD	A,ERR86		;program terminated
;
EXIT	AND	7FH		;make sure it returns
	LD	B,A		;pass number here
EXITB	LD	A,ERROR		;SVC #
	RST	SVC		;dislay the error
;
RETURN	LD	SP,0		;reset stack
STACK	EQU	$-2
	LD	HL,0		;get old break processor
OLDBRK	EQU	$-2
	LD	A,SETBRK	;SVC #
	RST	SVC		;restore old vector
	XOR	A		;return to caller ZERO
	RET
;
;	prompt for parameters
;
APARMS	LD	HL,PMSG		;prompt line
	LD	BC,1<8+79	;disp len + key len
	LD	DE,STRING	;input buffer
	LD	A,VIDKEY	;SVC #
	RST	SVC		;display & keyboard
;
	JR	NZ,EXIT		;display error
	JR	C,TERMIN	;break, terminate
	EX	DE,HL		;HL => input
	JR	SCAN		;parse params
;
	PAGE
;
;	checking out a drive
;
CHKDRIV	LD	C,0		;fetch drive number
DRIVE	EQU	$-1
	CALL	LOCDCT		;load IY with DCT
	LD	A,DCHECK	;check drive
	RST	DISKIO		;check it out
	LD	A,ERR8		;not available
	JR	NZ,EXIT		;no good, restart
;
	LD	HL,0		;set error counter
	LD	(ERRORS),HL	;to zero
;
	CALL	SETUPS		;setup pointers
	JP	CONT2		;continue
;
	PAGE
;
;	$SETUPS	- setup information from DCT
;
SETUPS	LD	A,(IY+21)	;grans/cylinder
	LD	E,2		;set mask bit
	DEC	A		;less one
;
GPTSF	JR	Z,GOTGPT	;have the mask
	SLA	E		;move mask bit
	DEC	A		;less gran
	JR	GPTSF		;go again
GOTGPT	LD	A,E		;fetch mask
	LD	(TOPGRAN),A	;save top gran mask
;
;	check for sufficient memory
;
	LD	HL,(TOPMEM)	;fetch topmem
	LD	DE,(LOWMEM)	;fetch low memory
	XOR	A		;clear carry
	LD	C,A		;save here
	LD	A,ERR82		;insufficient memory
	SBC	HL,DE		;HL = free memory
	JP	C,EXIT		;pointers corrupt!
	LD	A,(IY+18)	;# sectors in dir
	ADD	A,3		;even page + gat+hit
	LD	B,A		;B = # pages
	SBC	HL,BC		;enough memory?
	JP	C,EXIT		;not enough
;
;	check for lowmem < program end
;
	LD	DE,(LOWMEM)	;get low memory
	LD	HL,ZZZZZ	;program end
	SBC	HL,DE		;compare 'em
	JR	C,MEMHAV	;DE is OK
	LD	DE,ZZZZZ	;else use program end
MEMHAV	INC	D		;next page
	LD	E,0		;start of page
	EX	DE,HL		;HL => buffer area
	LD	(GATBUFF),HL	;save gat pointer
	INC	H		;bump pointer
	LD	(HITBUFF),HL	;save hit pointer
	INC	H		;bump pointer
	LD	(DIRGAT),HL	;save directory pointer
	INC	H		;bump page
	LD	(DIRHIT),HL	;save hit pointer
	INC	H
	LD	(DIRBUF),HL	;pointer to files
;
	CALL	CLRGAT		;reset GAT table
	CALL	CLRHIT		;reset HIT table
;
;	all OK, read in the directory
;
	CALL	RDDIR		;read directory
;
;	compute # free grans
;
	LD	HL,(DIRGAT)	;start of buffer
	LD	B,(IY+16)	;fetch cylinder count
	LD	DE,0		;DE = free gran counter
;
FREELP	LD	C,8		;check 8 bits
	LD	A,(HL)		;get GAT byte
FREEL	RRA			;shift out a bit
	CALL	NC,BUMPDE	;bump DE if set
	DEC	C		;C = bit counter
	JR	NZ,FREEL	;go more
	INC	HL		;next track
	DJNZ	FREELP		;do all tracks
;
	LD	HL,FREEMSG	;message for ascii
	PUSH	HL		;save pointer
	LD	B,0		;bin>ascii
	LD	A,BINDEC	;SVC #
	RST	SVC		;load the string
;
;	fetch name and date
;
	LD	HL,(DIRGAT)	;get gat buffer
	LD	L,0D0H		;start of name
	LD	DE,NAMMSG	;string to put it
	LD	BC,8		;8 chars
	LDIR			;move it in
	LD	C,8		;8 more chars
	LD	DE,DATMSG	;date string
	LDIR			;move it in
;
	POP	HL		;get message ptr back
	JP	DISPLAY		;display message
;
CONT2	LD	HL,(DIRBUF)	;start of directory
	LD	(DIRPOS),HL	;save file entry start
	LD	A,2		;current real sector
	LD	(DIRSEC),A	;save it
	LD	A,(IY+18)	;fetch top sector
	SUB	2		;less 2
	LD	B,A		;here for looper
	LD	DE,20H		;offset between files
;
;	use bit 5 as a flag bit, reset 'em all
;
CK00	LD	C,8		;8 entries / sector
CK000	RES	5,(HL)		;reset it
	ADD	HL,DE		;point to next one
	DEC	C		;less sector counter
	JR	NZ,CK000	;finish this sector
	DJNZ	CK00		;do rest of directory
;
;	setup loop to check all files
;
CHKLP	LD	HL,0		;get directory position
DIRPOS	EQU	$-2
	BIT	4,(HL)		;active entry?
;
	PUSH	AF		;save the flag
	CALL	NZ,CHECKA	;checker A (GAT)
	POP	AF		;restore flag
;
	PUSH	AF		;put 'em back
	CALL	NZ,CHECKB	;checker B (HIT)
	POP	AF
;
	CALL	NZ,CHECKC	;checker C
;
	LD	HL,(DIRPOS)	;refetch position
	LD	BC,20H		;offset between
	ADD	HL,BC		;next entry
	LD	(DIRPOS),HL	;resave position
;
	LD	A,L		;see if new sector
	OR	A		;at relative 00?
	JR	NZ,CHKLP	;nope, continue
;
	LD	A,0		;fetch directory sector
DIRSEC	EQU	$-1
	INC	A		;bump it
	LD	(DIRSEC),A	;resave it
	CP	(IY+18)		;test for top sector
	JR	NZ,CHKLP	;nope, go more
;
	BIT	5,(IY+5)	;rigid drive?
	JR	NZ,GOTRIG	;yes, skip lockout
;
;	merge lockout table with created GAT
;
	LD	DE,(DIRGAT)	;start of directory
	LD	E,60H		;start of lockout table
	LD	HL,(GATBUFF)	;created table
	LD	B,(IY+16)	;get cylinder count
;
RIGLP	LD	A,(DE)		;get lockout byte
	OR	(HL)		;merge with GAT
	LD	(HL),A		;put in table
	INC	DE		;bump table pointers
	INC	HL
	DJNZ	RIGLP		;finish it off
;
GOTRIG	LD	HL,(DIRGAT)	;GAT from disk
	LD	DE,(GATBUFF)	;created GAT
	LD	B,(IY+16)	;cylinder count
	LD	C,0		;relative cylinder
;
CHECKD	LD	A,(DE)		;get a byte
	CP	(HL)		;same?
	CALL	NZ,CHECKDB	;nope, display it
	INC	DE		;bump pointers
	INC	HL
	INC	C		;bump relative cylinder
	DJNZ	CHECKD		;finish it off
;
;	check HIT table
;
	LD	HL,(DIRHIT)	;start of HIT
	LD	DE,(HITBUFF)	;created HIT
	LD	BC,0		;B=looper, C=rel byte
;
CHECKE	LD	A,(DE)		;get a byte
	CP	(HL)		;same?
	CALL	NZ,CHECKEB	;nope, display it
	INC	HL		;bump pointers
	INC	DE
	INC	C		;bump rel byte
	DJNZ	CHECKE		;do 'em all
;
;	check for extraneous files leftover
;
	LD	HL,(DIRBUF)	;start of directory
	LD	(DIRPOS),HL	;save position
	LD	A,2		;relative sector
	LD	(DIRSEC),A	;save it
;
CHECK3	BIT	4,(HL)		;active entry?
	CALL	NZ,LEFTOVER	;yes, show it
	LD	BC,20H		;offset between
	ADD	HL,BC		;next entry
	LD	(DIRPOS),HL	;update position
;
	LD	A,L		;see if new sector
	OR	A		;yes?
	JR	NZ,CHECK3	;nope, continue
;
	LD	A,(DIRSEC)	;fetch dir sector
	INC	A		;bump it
	LD	(DIRSEC),A	;put it back
	CP	(IY+18)		;test for end
	JR	NZ,CHECK3	;nope, continue
	JP	COMPLETE	;program completed
;
;	active file, see if checked
;
LEFTOVER BIT	5,(HL)		;flag bit set?
	RET	NZ		;yes, it's OK
;
	LD	A,ERR101	;extension record no asgn
	JP	SHOWER		;display it & count
;
	PAGE
;
;	$CHECKB	- check HIT table
;
CHECKB	LD	HL,(DIRPOS)	;get record pointer
	BIT	7,(HL)		;extension?
	RET	NZ		;skip if yes
;
	PUSH	HL		;save pointer
	POP	IX		;pass it here
	LD	DE,(DIRBUF)	;start of file records
	OR	A		;clear carry
	SBC	HL,DE		;compute relative HIT
	LD	A,H		;fetch relative sector
	OR	L		;or with relative byte
	LD	(DEC),A		;save directory entry cod
;
	CALL	HASH		;compute HASH for file
	LD	(HIT),A		;save it
;
CHECKA1	LD	HL,(DIRHIT)	;HIT table
	LD	L,0		;get directory entry code
DEC	EQU	$-1
;
	LD	A,(HL)		;get the byte
	OR	A		;check for nil hit
	JR	Z,ZERHIT	;yes, display it
	LD	A,0		;get real HIT byte
HIT	EQU	$-1
	CP	(HL)		;same?
	JR	Z,HITOK		;yes, continue
;
	LD	A,ERR103	;invalid HIT byte
	JR	HITERR		;display it
;
ZERHIT	LD	A,ERR104	;zero hit byte
;
HITERR	CALL	SHOWER		;display the error
;
HITOK	LD	HL,(HITBUFF)	;computed table
	LD	A,(DEC)		;get DEC
	LD	L,A		;HL => relative byte
	LD	A,(HIT)		;get HIT
	LD	(HL),A		;save actual byte
;
	LD	IX,(DIRPOS)	;get current position
;
;	check for any extended entries
;
HITAA	LD	B,5		;5 extent areas max
;
HITA	LD	A,(IX+22)	;get extent byte
	CP	0FFH		;terminator?
	RET	Z		;yes, return
	CP	0FEH		;extension?
	JR	Z,HITB		;yes, check it out
;
	INC	IX		;next extent
	INC	IX
	DJNZ	HITA		;go more
	RET			;done with this record
;
HITB	LD	A,(IX+23)	;get extension DEC
	AND	0E0H		;compute relative byte
	LD	C,A		;save here
	LD	A,(IX+23)	;get again
	LD	(DEC),A		;save it
	AND	1FH		;get relative sector
	LD	B,A		;save here, BC = offset
	LD	HL,(DIRBUF)	;start of records
	ADD	HL,BC		;point to record
	PUSH	HL		;pass it
	POP	IX		;to IX for next loop
;
	LD	HL,(DIRHIT)	;HIT table
	LD	A,(DEC)		;get entry code
	LD	L,A		;HL => relative byte
	LD	A,(HL)		;get it
	OR	A		;zero?
	JR	Z,DOZERHIT	;zero extension
	LD	A,(HIT)		;get hit
	CP	(HL)		;same?
	JR	Z,HITOKA	;yes, continue
;
	LD	A,ERR105	;extended invalid hit
	JR	DOERHIT		;post the error
;
DOZERHIT LD	A,ERR106	;extended zero hit
;
DOERHIT	CALL	SHOWER		;display the error
;
HITOKA	LD	HL,(HITBUFF)	;created table
	LD	A,(DEC)		;get entry code
	LD	L,A		;HL => rel byte
	LD	A,(HIT)		;get real HIT
	LD	(HL),A		;save in created table
	JP	HITAA		;continue with extension
;
	PAGE
;
;	$CHECKA	- check GAT table
;
CHECKA	LD	HL,(DIRPOS)	;get position
	BIT	7,(HL)		;extended record?
	RET	NZ		;yes, skip this file
;
	SET	5,(HL)		;flag as done
	PUSH	HL		;pass pointer
	POP	IX		;to IX
	LD	B,11		;check filename area
;
CK1A	LD	A,(IX+5)	;get a byte
	CALL	VALCHR		;valid file char?
	JR	Z,CK1B		;nope, invalid
;
	INC	IX		;bump to next char
	DJNZ	CK1A		;finish it off
	JR	CK1C		;check next function
;
CK1B	LD	A,ERR19		;invalid filespec
	CALL	SHOWER		;display it
;
CK1C	LD	IX,(DIRPOS)	;next record
	PUSH	IX		;pass to HL
	POP	HL
	LD	DE,(DIRBUF)	;start of names
	OR	A		;clear carry
	SBC	HL,DE		;compute DEC
	LD	A,H		;fetch DEC into A
	OR	L
	LD	(DEC),A		;save it
	LD	BC,0		;for file positioning
	LD	(RELSEC),BC	;save relative sector
;
CK1D	XOR	A		;set rel extent to 0
	LD	(CKEXT),A	;save it
;
CK1E	LD	A,(IX+22)	;get extent byte
	CP	0FFH		;terminator?
	JP	Z,DONEWITH	;yes, done with it
	CP	0FEH		;extension?
	JR	Z,NXFIL		;yes, get it
;
	LD	D,A		;save rel track
	LD	E,0		;start sector
	LD	C,(IY+20)	;sectors / gran
	LD	A,(IX+23)	;get extent
	AND	0E0H		;start grans bit
	SRL	A
	SRL	A
	SRL	A
	SRL	A
	SRL	A		;align the bits
GNLP	OR	A		;have it?
	JR	Z,STGN		;have it
	PUSH	AF		;save
	LD	A,E		;get current sector
	ADD	A,C		;add sect/gran
	LD	E,A		;put it back
	POP	AF		;get counter
	DEC	A		;less one
	JR	GNLP		;try again
;
STGN	LD	A,(IX+23)	;get extent
	AND	1FH		;number of grans
	INC	A		;make it real
	LD	C,A		;save here
	CALL	SHOPRT		;check it out
;
	INC	IX		;bump to next element
	INC	IX
;
	LD	A,0		;see if end of extents
CKEXT	EQU	$-1
	INC	A		;bump it
	LD	(CKEXT),A	;put it back
	CP	5		;used 5 now?
	JR	NZ,CK1E		;nope, continue
;
	LD	A,ERR92		;no extent terminator
	JP	SHOWER		;display it
;
NXFIL	LD	A,(CKEXT)	;get extent
	CP	4		;last one?
	LD	A,ERR100	;assigned too early!
	CALL	NZ,SHOWER	;display error if true
;
	LD	A,(IX+23)	;get element back
	PUSH	AF		;save it
	AND	1FH		;relative sector
	LD	B,A		;save here
	POP	AF		;get back
	AND	0E0H		;relative byte
	LD	C,A		;BC = extension offset
	LD	HL,(DIRBUF)	;start of records
	ADD	HL,BC		;HL => extension
;
	PUSH	HL		;save it
	SET	5,(HL)		;set tested
	BIT	4,(HL)		;active entry?
	LD	A,ERR95		;forward link bad
	JP	Z,SHOWER-1	;if yes, abort file
;
	BIT	7,(HL)		;really an extension?
	LD	A,ERR96		;link->non-extension
	JP	Z,SHOWER-1	;error, abort file
;
	LD	A,(IX+23)	;get DEC
	POP	IX		;load IX with file record
	PUSH	AF		;save DEC
	LD	A,(DEC)		;fetch last DEC
	CP	(IX+1)		;backward link OK?
	LD	A,ERR93		;does not link back
	JP	NZ,SHOWER-1	;nope, abort file
	POP	AF		;restore DEC
;
	LD	(DEC),A		;save extension DEC
	JP	CK1D		;check it out
;
DONEWITH PUSH	HL		;save current pointer
	PUSH	DE		;save more
	PUSH	IX		;save too
	LD	IX,(DIRPOS)	;get current file
	LD	E,(IX+20)	;get end of file sector
	LD	D,(IX+21)
	LD	A,D		;nil?
	OR	E
	JR	NZ,DWTH		;yes, skip it
;
	LD	A,(IX+3)	;get end of file byte
	OR	A		;zero?
	JR	Z,DWTHY		;yes, go
;
DWTH	LD	HL,(RELSEC)	;# sectors allocated
	SBC	HL,DE		;compare
;
DWTHY	POP	IX
	POP	DE		;restore stack
	POP	HL
	RET	NC		;OK, return
;
;	end of file beyond assigned sectors
;
	LD	A,ERR99		;end of file past alloc.
	JP	SHOWER		;post it
;
SHOPRT	PUSH	HL		;save registers
	PUSH	DE
	PUSH	BC
;
PRTRLP	LD	B,(IY+20)	;sectors / gran
;
PRTLP	CALL	PRTHIS		;this extent
	JR	C,SKPFRS	;error, go!
;
	CALL	NEXSEC		;bump to next sector
	DJNZ	PRTLP		;finish this block off
	DEC	C
	JR	NZ,PRTRLP
;
SKPFRS	POP	BC		;unstack
	POP	DE
	POP	HL
	RET
;
PRTHIS	PUSH	HL		;save pointer
	LD	HL,0		;get relative sector
RELSEC	EQU	$-2
	INC	HL		;bump it
	LD	(RELSEC),HL	;put it back
	POP	HL		;restore pointer
;
	LD	A,(IY+16)	;cylinder count
	CP	D		;past disk boundary?
	RET	NC		;nope, return
;
	LD	A,ERR94		;beyond disk boundary
	CALL	SHOWER		;post it
	SCF			;set error
	RET			;return
;
	PAGE
;
;	$SHOWERR - post error type
;
	POP	DE		;fixup the stack
;
SHOWER	CALL	REGSAV		;save registers
;
	LD	(ERRSV),A	;save error number
;
	LD	HL,FILMSG	;string for filename
	LD	DE,FILMSG+1	;clear it out
	LD	BC,20		;20 chars long
	LD	(HL),20H	;put a blank
	LDIR			;put all blanks
;
	LD	IX,(DIRPOS)	;get file pointer
	LD	A,(IX)		;get file type
	LD	(SHOPAS),A	;save it for display
	LD	DE,FILMSG	;point to message
	LD	B,8		;8 chars in name
;
SHO1	LD	A,(IX+5)	;get name byte
	CP	' '		;a char here?
	JR	Z,SHO2		;not part of name
;
	CALL	VALCHR		;valid filename char?
	CALL	Z,FIXSHO	;put a block here
	LD	(DE),A		;put into string
;
	INC	IX		;bump pointers
	INC	DE
	DJNZ	SHO1		;finish it off
	LD	A,(IX+5)	;check for any extension
	CP	' '		;any?
	JR	Z,SHO4		;nope, go
;
SHO2	LD	B,3		;3 char extension
	LD	A,'/'		;put into name
	LD	(DE),A		;to the string
	INC	DE		;bump pointer
;
	LD	IX,(DIRPOS)	;fetch start
;
SHO3	LD	A,(IX+13)	;get ext char
	CP	' '		;done?
	JR	Z,SHO4		;yes, go
;
	CALL	VALCHR		;valid char?
	CALL	Z,FIXSHO	;put a block if no
	LD	(DE),A		;put into string
	INC	DE		;bump pointer
	INC	IX		;bump filename ptr
	DJNZ	SHO3		;finish it off
;
SHO4	LD	B,0		;get attribute byte
SHOPAS	EQU	$-1
	INC	DE		;bump string pointer
;
	LD	A,'S'		;system file?
	BIT	6,B		;yes?
	JR	Z,SHO5		;no, go
	LD	(DE),A		;else put in buffer
	INC	DE		;bump pointer
;
SHO5	LD	A,'I'		;invisible
	BIT	3,B		;yes?
	JR	Z,SHO6		;no, go
	LD	(DE),A		;else put in buffer
	INC	DE		;bump it
;
SHO6	LD	A,B		;get byte
	AND	7		;get prot level
	JR	Z,SHO8		;0, go
	PUSH	AF		;save it
	LD	A,'P'		;set protected
	LD	(DE),A		;to the string
	INC	DE		;bump it
	LD	A,'='		;for numeric
	LD	(DE),A		;to the string
	INC	DE		;bump it
	POP	AF		;restore value
	ADD	A,'0'		;make it ascii
	LD	(DE),A		;to the string
	INC	DE		;bump it
;
SHO8	LD	HL,(ERRORS)	;get number of errors
	INC	HL		;bump it
	LD	(ERRORS),HL	;put it back
;
	LD	A,(DIRPOS)	;directory position
	LD	HL,WHR2		;string where it goes
	CALL	HEXCV		;make it hex
;
	LD	A,(DIRSEC)	;get directory sector
	LD	HL,WHR1		;text pointer
	CALL	ASCII		;make it decimal
;
	LD	HL,FILMSG	;location message
	CALL	DISPLAY		;dual display
;
SPSHOW	LD	B,0		;get error number
ERRSV	EQU	$-1
	LD	A,ERROR		;SVC #
	RST	SVC		;display it
	JP	PAUSE		;check for pause
;
FIXSHO	LD	A,5FH		;char to replace with
	RET
;
	PAGE
;
;	check a file
;
CHKFILE	INC	DE		;DCB for file
	EX	DE,HL		;swap 'em
	CALL	DISPLAY		;display it
	LD	HL,SPCS		;then a gap
	CALL	DISPLAY		;display it too
;
	LD	DE,DCB+1	;get it back
	LD	HL,OLIST	;open list
	LD	A,OPEN		;SVC #
	RST	SVC		;open it up
	JP	NZ,EXIT		;nope, cancel
;
	PUSH	DE		;pass to IX
	POP	IX		;IX => open DCB
	LD	A,(IX+16)	;get device number
	LD	(DRIVE),A	;save it
	LD	C,A		;need it here
	CALL	LOCDCT		;get drive DCT
;
	LD	A,(IX+17)	;get DEC
	LD	(DEC),A		;save it
	LD	HL,MXB		;text string
	CALL	HEXCV		;make it hex
;
	LD	A,(IX+11)	;end of file byte
	LD	HL,MXC		;string
	CALL	HEXCV		;to hex ascii
;
	LD	A,(IX+15)	;logical record length
	LD	HL,MXE		;string
	CALL	ASCII		;to decimal
;
	LD	E,(IX+14)	;end of file sector
	LD	D,(IX+13)	;next MSB
	LD	HL,MXD		;dest string
	LD	B,0		;direction
	LD	A,BINDEC	;binary => decimal
	RST	SVC		;load string
;
	LD	HL,MX0		;point to string
	CALL	DISPLAY		;dual display
;
	CALL	SETUPS		;setup pointers
	LD	A,(DEC)		;get DEC
	AND	0E0H		;relative position
	LD	C,A		;pass here
	LD	A,(DEC)		;get again
	AND	1FH		;relative sector
	LD	B,A		;BC = entry offset
	ADD	A,2		;adjust to actual sector
	LD	(DIRSEC),A	;save it
	LD	HL,(DIRBUF)	;start of records
	ADD	HL,BC		;HL => directory record
	LD	(DIRPOS),HL	;save it
;
	CALL	CHECKA		;checker A
	CALL	CHECKB		;checker B
	CALL	CHECKC		;checker C
;
	JP	COMPLETE	;finished
;
	PAGE
;
;	clear out the created HIT table
;
CLRHIT	LD	HL,(HITBUFF)	;start of buffer
	LD	D,H		;pass to DE
	LD	E,L
	INC	DE		;DE = HITBUFF+1
	LD	BC,0FFH		;length to fill
	LD	(HL),0		;with zeroes
	LDIR			;do 'em all
	RET			;done
;
	PAGE
;
;	clear out the created GAT table
;
CLRGAT	LD	HL,(GATBUFF)	;start of GAT
	LD	D,H		;pass to DE
	LD	E,L
	INC	DE		;DE = GATBUFF+1
	LD	BC,0FFH		;length
	LD	(HL),0FFH	;set all nil
	LDIR			;do 'em all
;
	LD	B,(IY+21)	;grans / cylinder
	LD	C,8		;bits / byte
;
	XOR	A		;start bit pattern
NXNY	SCF			;set a bit
	RRA			;move it in
	DEC	C		;less counter
	DJNZ	NXNY		;less grans / track
	DEC	C		;see if done
	INC	C
	JR	Z,NXNZ		;yes, go!
;
NXNU	OR	A		;clear carry
	RRA			;align bits
	DEC	C		;continue
	JR	NZ,NXNU		;for rest of mask
;
NXNZ	LD	IX,(GATBUFF)	;start of gat
	CPL			;reverse pattern
	LD	B,(IY+16)	;cylinder count
	BIT	5,(IY+5)	;rigid drive?
	JR	NZ,NXNT		;yes, no lockout
;
NXNV	LD	(IX),A		;put in the byte
	LD	(IX+60H),A	;to lockout table too
	INC	IX		;bump pointer
	DJNZ	NXNV		;do whole GAT
	RET			;done, return
;
NXNT	LD	(IX),A		;put in GAT byte
	INC	IX		;next cylinder
	DJNZ	NXNT		;do whole GAT
	RET			;done, return
;
PAUSE	LD	A,KBCHAR	;keyboard character
	RST	SVC		;fetch one
	RET	NZ		;nope, return
;
	LD	A,B		;fetch the char
	CP	3		;break?
	JP	Z,TERMIN	;yes, abort
;
	OR	A		;hold?
	RET	NZ		;nope, return
;
PAUWT	LD	A,KBCHAR	;keyboard
	RST	SVC		;get a key
	JR	NZ,PAUWT	;wait for one
;
	LD	A,B		;get char
	CP	3		;break?
	JP	Z,TERMIN	;abort if yes
	OR	A		;hold?
	JR	NZ,PAUWT	;nope, wait more
	RET			;back
;
;	advance DE to next sector
;
NEXSEC	INC	E		;bump sector
	LD	A,(IY+22)	;fetch top sector
	SUB	E		;at end?
	RET	NZ		;nope, cancel
;
	LD	E,A		;reset to 0
	LD	A,D		;get track
	CP	(IY+16)		;end of disk?
	RET	NC		;yes, don't bump it
	INC	D		;else advance track
	RET			;done
;
	PAGE
;
;	function completed
;
COMPLETE LD	DE,0		;fetch # of errors
ERRORS	EQU	$-2
	LD	HL,ERROR0	;where it goes
	LD	B,0		;command direction
	LD	A,BINDEC	;binary to decimal
	RST	SVC		;load string
;
	LD	HL,TOTAL	;point to string
	CALL	DISPLAY		;dual display
;
	JP	RETURN		;exit program
;
	PAGE
;
;	read the directory
;
RDDIR	LD	A,(DRIVE)	;fetch the drive number
	LD	C,A		;pass it here
	LD	B,(IY+18)	;get number of sectors
	LD	E,0		;starting sector
	LD	HL,(DIRGAT)	;where to load data
;
RDDLP	PUSH	BC		;save drive, counter
	LD	B,0		;for sector offset
	LD	A,SREAD		;read directory sector
	RST	DISKIO		;read the sector
	POP	BC		;unstack
;
	JP	NZ,EXIT		;error!
;
	INC	H		;bump buffer pointer
	INC	E		;bump sector
	DJNZ	RDDLP		;finish it off
	XOR	A		;set Z
	RET			;back
;
;	convert to hex ascii
;
HEXCV	LD	B,A		;save it
	RRCA			;align to low nibble
	RRCA
	RRCA
	RRCA
	CALL	HEXTST		;make it ascii
	LD	A,B		;fetch LSB
;
HEXTST	AND	0FH		;mask upper bits
	ADD	A,'0'		;add the ascii
	CP	'9'+1		;out of numeric range
	JR	C,HEXPUT	;put in string
	ADD	A,7		;adjust for A-F
HEXPUT	LD	(HL),A		;put in string
	INC	HL		;bump pointer
	RET			;done
;
;	binary to decimal ascii
;
ASCII	LD	C,'0'		;start digit
ASC1	SUB	100		;less hundred
	JR	C,ASC2		;have digit
	INC	C
	JR	ASC1
ASC2	ADD	A,100		;add it back
	LD	(HL),C		;insert digit
	INC	HL		;bump pointer
	LD	C,'0'		;start again
ASC3	SUB	10		;less 10's
	JR	C,ASC4
	INC	C		;bump ascii
	JR	ASC3		;go more
ASC4	LD	(HL),C		;NSB
	INC	HL
	ADD	A,3AH		;ascii + last sub
	LD	(HL),A		;to the string
	RET			;done
;
	PAGE
;
BUMPDE	INC	DE		;bumper for DE
	RET
;
;	compute HASH code
;
HASH	PUSH	IX		;save pointer
	PUSH	BC		;need for compute
	LD	C,0		;start hash
	LD	B,0BH		;# chars to hash
;
HASHL	LD	A,(IX+5)	;fetch a byte
	INC	IX		;bump pointer
	XOR	C		;'hash' it
	RLCA
	LD	C,A
	DJNZ	HASHL		;finish it
;
	OR	A		;see if 0
	JR	NZ,HASHOK	;nope, return
	INC	A		;make it one
;
HASHOK	POP	BC		;unstack & return
	POP	IX
	RET
;
	PAGE
;
;	checker C
;
CHECKC	LD	HL,(DIRPOS)	;current file ptr
	BIT	7,(HL)		;extension?
	RET	NZ		;yes, return
;
	PUSH	HL		;pass to IX
	POP	IX
	XOR	A		;set check flag
	LD	(CFLAG),A	;to nil
;
CHECKC1	LD	B,5		;5 extents per entry
;
CHECKC2	LD	A,(IX+22)	;get an extent
;
	CP	0FFH		;terminator?
	RET	Z		;yes, done
;
	CP	0FEH		;extension?
	JP	Z,CHECKC3	;yes, fetch it
;
	LD	D,A		;save track
	LD	A,(IX+23)	;get extent
	AND	1FH		;# grans
	INC	A		;adjust to real
	LD	C,A		;save here
	LD	A,(IX+23)	;get again
	LD	E,1		;bit mask
	RRCA			;align the bits
	RRCA
	RRCA
	RRCA
	RRCA
	AND	7		;want top 3 only
;
FINGLP	JR	Z,FINGHV	;have the gran
	SLA	E		;shift the bit over
	DEC	A		;continue
	JR	FINGLP
;
FINGHV	LD	HL,(GATBUFF)	;GAT start
	LD	L,D		;D = track
;
RELGRNS	LD	A,(HL)		;get a byte
	AND	E		;mask with gran bits
	JR	Z,RELNXT	;go if done
;
	LD	A,0		;check the flag
CFLAG	EQU	$-1
	OR	A		;used?
	JR	NZ,RELNXT	;no, go
;
	INC	A		;else reset it
	LD	(CFLAG),A	;don't display error
	LD	A,ERR102	;multiple files > gran
	CALL	SHOWER		;display it
;
RELNXT	LD	A,(HL)		;fix the byte
	OR	E
	LD	(HL),A		;put it back
	DEC	C		;more to do?
	JR	Z,RELGDNE	;nope, done
	LD	A,0		;get the top gran
TOPGRAN	EQU	$-1
	SLA	E		;move test bit over
	CP	E		;done with cylinder?
	JR	NZ,RELGRNS	;nope, continue
;
	INC	L		;next cylinder
	LD	E,1		;reset the mask
	JR	RELGRNS		;go more
;
RELGDNE	INC	IX		;next extent
	INC	IX
	DJNZ	CHECKC2		;finish it off
	RET			;done
;
CHECKC3	LD	A,(IX+23)	;get DEC for extension
	AND	0E0H		;relative byte
	LD	C,A		;save it
	LD	A,(IX+23)	;get again
	AND	1FH		;sector offset
	LD	B,A		;BC = file offset
	LD	HL,(DIRBUF)	;start of records
	ADD	HL,BC		;HL => name
	PUSH	HL		;pass to IX
	POP	IX
	JP	CHECKC1		;check this out
;
CHECKDB	PUSH	HL		;save 'em
	PUSH	DE
	PUSH	BC
	LD	A,C		;get relative byte
	LD	HL,ERRMAN	;string pointer
	CALL	ASCII		;make it ascii
	BIT	5,(IY+5)	;rigid?
	JR	NZ,GODBCK	;yes, go
;
	LD	BC,60H		;offset to lockout
	ADD	HL,BC		;point to it
	LD	A,(HL)		;fetch it
	INC	A		;locked out?
	JR	Z,OKBCK		;else it is locked
;
GODBCK	LD	HL,ERRMA	;error message
	JR	CONTCKB		;continue
;
CHECKEB	PUSH	HL		;save 'em
	PUSH	DE
	PUSH	BC
;
	LD	A,C		;get rel byte
	LD	HL,ERRMBN	;text pointer
	CALL	HEXCV		;to ascii
	LD	HL,ERRMB	;point to it
;
CONTCKB	CALL	DISPLAY		;display it
	LD	HL,(ERRORS)	;get error count
	INC	HL		;bump it
	LD	(ERRORS),HL	;update
;
OKBCK	POP	BC		;unstack & return
	POP	DE
	POP	HL
	RET
;
	PAGE
;
;	display string
;
DISPLAY	PUSH	HL		;this too
DISPLP	LD	A,(HL)		;get a char
	CP	3		;terminator?
	JR	Z,DISPDN
	CALL	DSPCHR		;display character
	INC	HL		;bump pointer
	JR	DISPLP		;do next one
;
DISPDN	POP	HL		;unstack
	JP	PAUSE		;check for pause
;
DSPCHR	PUSH	BC		;save it
	LD	B,A		;pass character
	LD	A,VDCHAR	;SVC #
	RST	SVC		;send the byte
	POP	BC		;restore stack
	RET	Z		;OK
	JP	EXIT		;report error, abort
;
	PAGE
;
;	check for valid character
;
VALCHR	PUSH	HL		;save
	PUSH	BC
	LD	HL,VALTBL	;table of chars
	LD	BC,18		;18 here
	CPIR
	POP	BC
	POP	HL
	RET
;
VALTBL	DEFM	'!'
	DEFM	'*'
	DEFM	'?'
	DEFM	'('
	DEFM	')'
	DEFM	','
	DEFM	'.'
	DEFM	'/'
	DEFM	';'
	DEFM	':'
	DEFM	'@'
	DEFM	'='
	DEFM	'"'
	DEFB	27H
	DEFB	3
	DEFB	13
	DEFB	7BH
	DEFB	7DH
;##
	PAGE
;
;	text
;
ERRMA	DEFM	'Cylinder '
ERRMAN	DEFM	'xxx Has An Invalid GAT Table Byte'
	DEFB	13
	DEFB	3
;
ERRMB	DEFM	'HIT Byte at '
ERRMBN	DEFM	'xxH invalid or extraneous'
	DEFB	13
	DEFB	3
;
FREEMSG	DEFM	'xxxxx Free Grans, Name = '
NAMMSG	DEFM	'xxxxxxxx, Date = '
DATMSG	DEFM	'xxxxxxxx'
	DEFB	13
	DEFB	13
	DEFB	3
;
TOTAL	DEFB	13
ERROR0	DEFM	'xxxxx Total Errors'
	DEFB	13
	DEFB	3
;
FILMSG	DEFM	'xxxxxxxxxxxxxxxxxxxxx '
	DEFM	'@ Directory Sector '
WHR1	DEFM	'xxx, Relative Byte '
WHR2	DEFM	'xxH'
	DEFB	13
	DEFB	3
;
SPCS	DEFM	' - '
	DEFB	3
;
MX0	DEFM	'DEC='
MXB	DEFM	'xxH, EOFB='
MXC	DEFM	'xxH, LRL='
MXE	DEFM	'xxx, EOFS='
MXD	DEFM	'xxxxx'
	DEFB	13
	DEFB	3
;
HELLO	DEFB	13
	DEFM	'DIRCHECK - '
	DEFM	'DosPLUS II Directory Check Utility - '
	DEFM	'Series A.00'
	DEFB	13
	DEFM	'(c)(p) Copyright 1982 '
	DEFM	'by MicroPower, Inc.'
	DEFB	13
	DEFB	13
	DEFB	13
	DEFB	3
PMSG	DEFB	'*'
OLIST	DEFW	BUFFER
	DEFW	BUFFER
	DEFW	0000H
	DEFB	'R'
	DEFB	0
	DEFB	'F'
	DEFB	0
	DEFB	0
;
CHKBLK	DEFB	0		;work area
	DEFW	DCB		;source
	DEFW	DCB		;dest
	DEFW	BUFFER		;dummy for mask
	DEFW	PARAMS		;param block
;
PARAMS	DEFB	0		;nil params
;
GATBUFF	DEFW	0		;created GAT table
HITBUFF	DEFW	0		;created HIT table
DIRGAT	DEFW	0		;disk GAT table
DIRHIT	DEFW	0		;disk HIT table
DIRBUF	DEFW	0		;disk file area
;
DCB	DEFS	41		;DCB area
;
ZZZZZ	EQU	$
;
	END	ENTRY
